/*-
 * suff.c --
 *	Functions to maintain suffix lists and find implicit dependents
 *	using suffix transformation rules
 *
 * Copyright (c) 1988 by the Regents of the University of California
 * Copyright (c) 1988 by Adam de Boor
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appears in all copies.  Neither the University of California
 * nor Adam de Boor makes any representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Interface:
 *	Suff_Init 	    	Initialize all things to do with suffixes.
 *
 *	Suff_DoPaths	    	This function is used to make life easier
 *	    	  	    	when searching for a file according to its
 *	    	  	    	suffix. It takes the global search path,
 *	    	  	    	as defined using the .PATH: target, and appends
 *	    	  	    	its directories to the path of each of the
 *	    	  	    	defined suffixes, as specified using
 *	    	  	    	.PATH<suffix>: targets. In addition, all
 *	    	  	    	directories given for suffixes labeled as
 *	    	  	    	include files or libraries, using the .INCLUDES
 *	    	  	    	or .LIBS targets, are played with using
 *	    	  	    	Dir_MakeFlags to create the .INCLUDES and
 *	    	  	    	.LIBS global variables.
 *
 *	Suff_ClearSuffixes  	Clear out all the suffixes and defined
 *	    	  	    	transformations.
 *
 *	Suff_IsTransform    	Return TRUE if the passed string is the lhs
 *	    	  	    	of a transformation rule.
 *
 *	Suff_AddSuffix	    	Add the passed string as another known suffix.
 *
 *	Suff_GetPath	    	Return the search path for the given suffix.
 *
 *	Suff_AddInclude	    	Mark the given suffix as denoting an include
 *	    	  	    	file.
 *
 *	Suff_AddLib	    	Mark the given suffix as denoting a library.
 *
 *	Suff_AddTransform   	Add another transformation to the suffix
 *	    	  	    	graph. Returns  GNode suitable for framing, I
 *	    	  	    	mean, tacking commands, attributes, etc. on.
 *
 *	Suff_SetNull	    	Define the suffix to consider the suffix of
 *	    	  	    	any file that doesn't have a known one.
 *
 *	Suff_FindDeps	    	Find implicit sources for and the location of
 *	    	  	    	a target based on its suffix. Returns the
 *	    	  	    	bottom-most node added to the graph or NILGNODE
 *	    	  	    	if the target had no implicit sources.
 */
#ifndef lint
static char     *rcsid = "$Id: suff.c,v 1.57 89/01/26 12:16:54 adam Exp $ SPRITE (Berkeley)";
#endif lint

#include    	  <stdio.h>
#include	  "make.h"
#include    	  "bit.h"

static Lst       sufflist;	/* Lst of suffixes */
static Lst       transforms;	/* Lst of transformation rules */

static int        sNum = 0;	/* Counter for assigning suffix numbers */

/*
 * Structure describing an individual suffix.
 */
typedef struct {
    char         *name;	    	/* The suffix itself */
    int		 nameLen;	/* Length of the suffix */
    short	 flags;      	/* Type of suffix */
#define SUFF_INCLUDE	  0x01	    /* One which is #include'd */
#define SUFF_LIBRARY	  0x02	    /* One which contains a library */
#define SUFF_NULL 	  0x04	    /* The empty suffix */
    Lst    	 searchPath;	/* The path along which files of this suffix
				 * may be found */
    int          sNum;	      	/* The suffix number */
    Lst          parents;	/* Suffixes we have a transformation to */
    Lst          children;	/* Suffixes we have a transformation from */
} Suff;

/*
 * Structure used in the search for implied sources.
 */
typedef struct _Src {
    char           *file;	/* The file to look for */
    Suff           *suff;	/* The suffix on the file */
    struct _Src    *parent;	/* The Src for which this is a source */
    GNode          *node;	/* The node describing the file */
} Src;

static char 	  *pref;      	/* Common prefix between two sides of
				 * a transformation rule */

static Suff 	  *suffNull;    /* The NULL suffix for this run */


 	/*************** Lst Predicates ****************/
/*-
 *-----------------------------------------------------------------------
 * SuffStrIsPrefix  --
 *	See if pref is a prefix of str.
 *
 * Results:
 *	NULL if it ain't, pointer to character in str after prefix if so
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static char    *
SuffStrIsPrefix (pref, str)
    register char  *pref;	/* possible prefix */
    register char  *str;	/* string to check */
{
    while (*str && *pref == *str) {
	pref++;
	str++;
    }

    return (*pref ? NULL : str);
}

/*-
 *-----------------------------------------------------------------------
 * SuffSuffIsSuffix  --
 *	See if suff is a suffix of str. Str should point to THE END of the
 *	string to check. (THE END == the null byte)
 *
 * Results:
 *	NULL if it ain't, pointer to character in str before suffix if
 *	it is.
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static char *
SuffSuffIsSuffix (s, str)
    register Suff  *s;		/* possible suffix */
    char           *str;	/* string to examine */
{
    register char  *p1;	    	/* Pointer into suffix name */
    register char  *p2;	    	/* Pointer into string being examined */

    p1 = s->name + s->nameLen;
    p2 = str;

    while (p1 >= s->name && *p1 == *p2) {
	p1--;
	p2--;
    }

    return (p1 == s->name - 1 ? p2 : NULL);
}

/*-
 *-----------------------------------------------------------------------
 * SuffSuffIsSuffixP --
 *	Predicate form of SuffSuffIsSuffix. Passed as the callback function
 *	to Lst_Find.
 *
 * Results:
 *	0 if the suffix is the one desired, non-zero if not.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------
 */
SuffSuffIsSuffixP(s, str)
    Suff    	*s;
    char    	*str;
{
    return(!SuffSuffIsSuffix(s, str));
}

/*-
 *-----------------------------------------------------------------------
 * SuffSuffHasName --
 *	Callback procedure for finding a suffix based on its name. Used by
 *	Suff_GetPath.
 *
 * Results:
 *	0 if the suffix is of the given name. non-zero otherwise.
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static int
SuffSuffHasName (s, sname)
    Suff    *s;	    	    /* Suffix to check */
    char    *sname; 	    /* Desired name */
{
    return (strcmp (sname, s->name));
}

/*-
 *-----------------------------------------------------------------------
 * SuffSuffIsPrefix  --
 *	See if the suffix described by s is a prefix of the string. Care
 *	must be taken when using this to search for transformations and
 *	what-not, since there could well be two suffixes, one of which
 *	is a prefix of the other...
 *
 * Results:
 *	0 if s is a prefix of str. non-zero otherwise
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static int
SuffSuffIsPrefix (s, str)
    Suff           *s;		/* suffix to compare */
    char           *str;	/* string to examine */
{
    return (SuffStrIsPrefix (s->name, str) == NULL ? 1 : 0);
}

/*-
 *-----------------------------------------------------------------------
 * SuffGNHasName  --
 *	See if the graph node has the desired name
 *
 * Results:
 *	0 if it does. non-zero if it doesn't
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static int
SuffGNHasName (gn, name)
    GNode          *gn;		/* current node we're looking at */
    char           *name;	/* name we're looking for */
{
    return (strcmp (name, gn->name));
}

 	    /*********** Maintenance Functions ************/
/*-
 *-----------------------------------------------------------------------
 * SuffFree  --
 *	Free up all memory associated with the given suffix structure.
 *
 * Results:
 *	none
 *
 * Side Effects:
 *	the suffix entry is detroyed
 *-----------------------------------------------------------------------
 */
static void
SuffFree (s)
    Suff           *s;
{
    Lst_Destroy (s->children, NOFREE);
    Lst_Destroy (s->parents, NOFREE);
    Lst_Destroy (s->searchPath, Dir_Destroy);
    free ((Address)s->name);
    free ((Address)s);
}

/*-
 *-----------------------------------------------------------------------
 * SuffInsert  --
 *	Insert the suffix into the list keeping the list ordered by suffix
 *	numbers.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Not really
 *-----------------------------------------------------------------------
 */
static void
SuffInsert (l, s)
    Lst           l;		/* the list where in s should be inserted */
    Suff          *s;		/* the suffix to insert */
{
    LstNode 	  ln;		/* current element in l we're examining */
    Suff          *s2;		/* the suffix descriptor in this element */

    if (Lst_Open (l) == FAILURE) {
	return;
    }
    while ((ln = Lst_Next (l)) != NILLNODE) {
	s2 = (Suff *) Lst_Datum (ln);
	if (s2->sNum > s->sNum) {
	    break;
	}
    }

    Lst_Close (l);
    if (ln == NILLNODE) {
	(void)Lst_AtEnd (l, (ClientData)s);
    } else if (s2->sNum != s->sNum) {
	(void)Lst_Insert (l, ln, (ClientData)s);
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_ClearSuffixes --
 *	This is gross. Nuke the list of suffixes but keep all transformation
 *	rules around. The transformation graph is destroyed in this process,
 *	but we leave the list of rules so when a new graph is formed the rules
 *	will remain.
 *	This function is called from the parse module when a
 *	.SUFFIXES:\n line is encountered.
 *
 * Results:
 *	none
 *
 * Side Effects:
 *	the sufflist and its graph nodes are destroyed
 *-----------------------------------------------------------------------
 */
void
Suff_ClearSuffixes ()
{
    Lst_Destroy (sufflist, SuffFree);

    sufflist = Lst_Init(FALSE);
    sNum = 0;
    suffNull = (Suff *)NULL;
}

/*-
 *-----------------------------------------------------------------------
 * SuffParseTransform --
 *	Parse a transformation string to find its two component suffixes.
 *
 * Results:
 *	TRUE if the string is a valid transformation and FALSE otherwise.
 *
 * Side Effects:
 *	The passed pointers are overwritten.
 *
 *-----------------------------------------------------------------------
 */
static Boolean
SuffParseTransform(str, srcPtr, targPtr)
    char    	  	*str;	    	/* String being parsed */
    Suff    	  	**srcPtr;   	/* Place to store source of trans. */
    Suff    	  	**targPtr;  	/* Place to store target of trans. */
{
    register LstNode	srcLn;	/* element in suffix list of trans source */
    register Suff    	*src;	/* Source of transformation */
    register LstNode    targLn;	/* element in suffix list of trans target */
    register char    	*str2;	/* Extra pointer (potential target suffix) */

    srcLn = NILLNODE;
    
    /*
     * Loop looking first for a suffix that matches the start of the
     * string and then for one that exactly matches the rest of it. If
     * we can find two that meet these criteria, we've successfully
     * parsed the string.
     */
    while (1) {
	if (srcLn == NILLNODE) {
	    srcLn = Lst_Find(sufflist, (ClientData)str, SuffSuffIsPrefix);
	} else {
	    srcLn = Lst_FindFrom (sufflist, Lst_Succ(srcLn), (ClientData)str,
				  SuffSuffIsPrefix);
	}
	if (srcLn == NILLNODE) {
	    /*
	     * Ran out of source suffixes -- no such rule
	     */
	    return (FALSE);
	}
	src = (Suff *) Lst_Datum (srcLn);
	str2 = str + strlen (src->name);
	targLn = Lst_Find(sufflist, (ClientData)str2, SuffSuffHasName);
	if (targLn != NILLNODE) {
	    *srcPtr = src;
	    *targPtr = (Suff *)Lst_Datum(targLn);
	    return (TRUE);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_IsTransform  --
 *	Return TRUE if the given string is a transformation rule
 *
 *
 * Results:
 *	TRUE if the string is a concatenation of two known suffixes.
 *	FALSE otherwise
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
Boolean
Suff_IsTransform (str)
    char          *str;	    	/* string to check */
{
    Suff    	  *src, *targ;

    return (SuffParseTransform(str, &src, &targ));
}

/*-
 *-----------------------------------------------------------------------
 * Suff_AddTransform --
 *	Add the transformation rule described by the line to the
 *	list of rules and place the transformation itself in the graph
 *
 * Results:
 *	The node created for the transformation in the transforms list
 *
 * Side Effects:
 *	The node is placed on the end of the transforms Lst and links are
 *	made between the two suffixes mentioned in the target name
 *-----------------------------------------------------------------------
 */
GNode *
Suff_AddTransform (line)
    char          *line;	/* name of transformation to add */
{
    GNode         *gn;		/* GNode of transformation rule */
    Suff          *s,		/* source suffix */
                  *t;		/* target suffix */
    LstNode 	  ln;	    	/* Node for existing transformation */

    ln = Lst_Find (transforms, (ClientData)line, SuffGNHasName);
    if (ln == NILLNODE) {
	/*
	 * Make a new graph node for the transformation. It will be filled in
	 * by the Parse module. 
	 */
	gn = Targ_NewGN (line);
	(void)Lst_AtEnd (transforms, (ClientData)gn);
    } else {
	/*
	 * New specification for transformation rule. Just nuke the old list
	 * of commands so they can be filled in again... We don't actually
	 * free the commands themselves, because a given command can be
	 * attached to several different transformations.
	 */
	gn = (GNode *) Lst_Datum (ln);
	Lst_Destroy (gn->commands, NOFREE);
	gn->commands = Lst_Init (FALSE);
    }

    gn->type = OP_TRANSFORM;

    (void)SuffParseTransform(line, &s, &t);

    /*
     * link the two together in the proper relationship and order 
     */
    SuffInsert (t->children, s);
    SuffInsert (s->parents, t);

    return (gn);
}

/*-
 *-----------------------------------------------------------------------
 * SuffRebuildGraph --
 *	Called from Suff_AddSuffix via Lst_ForEach to search through the
 *	list of existing transformation rules and rebuild the transformation
 *	graph when it has been destroyed by Suff_ClearSuffixes. If the
 *	given rule is a transformation involving this suffix and another,
 *	existing suffix, the proper relationship is established between
 *	the two.
 *
 * Results:
 *	Always 0.
 *
 * Side Effects:
 *	The appropriate links will be made between this suffix and
 *	others if transformation rules exist for it.
 *
 *-----------------------------------------------------------------------
 */
static int
SuffRebuildGraph(transform, s)
    GNode   	  	*transform; /* Transformation to test */
    Suff    	  	*s;	    /* Suffix to rebuild */
{
    register char 	*cp;
    register LstNode	ln;
    register Suff  	*s2;

    /*
     * First see if it is a transformation from this suffix.
     */
    cp = SuffStrIsPrefix(s->name, transform->name);
    if (cp != (char *)NULL) {
	ln = Lst_Find(sufflist, (ClientData)cp, SuffSuffHasName);
	if (ln != NILLNODE) {
	    /*
	     * Found target. Link in and return, since it can't be anything
	     * else.
	     */
	    s2 = (Suff *)Lst_Datum(ln);
	    SuffInsert(s2->children, s);
	    SuffInsert(s->parents, s2);
	    return(0);
	}
    }

    /*
     * Not from, maybe to?
     */
    cp = SuffSuffIsSuffix(s, transform->name + strlen(transform->name));
    if (cp != (char *)NULL) {
	/*
	 * Null-terminate the source suffix in order to find it.
	 */
	cp[1] = '\0';
	ln = Lst_Find(sufflist, (ClientData)transform->name, SuffSuffHasName);
	/*
	 * Replace the start of the target suffix
	 */
	cp[1] = s->name[0];
	if (ln != NILLNODE) {
	    /*
	     * Found it -- establish the proper relationship
	     */
	    s2 = (Suff *)Lst_Datum(ln);
	    SuffInsert(s->children, s2);
	    SuffInsert(s2->parents, s);
	}
    }
    return(0);
}

/*-
 *-----------------------------------------------------------------------
 * Suff_AddSuffix --
 *	Add the suffix in string to the end of the list of known suffixes.
 *	Should we restructure the suffix graph? Make doesn't...
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	A GNode is created for the suffix and a Suff structure is created and
 *	added to the suffixes list unless the suffix was already known.
 *-----------------------------------------------------------------------
 */
void
Suff_AddSuffix (str)
    char          *str;	    /* the name of the suffix to add */
{
    Suff          *s;	    /* new suffix descriptor */
    LstNode 	  ln;

    ln = Lst_Find (sufflist, (ClientData)str, SuffSuffHasName);
    if (ln == NILLNODE) {
	s = (Suff *) malloc (sizeof (Suff));

	s->name =   	Str_New (str);
	s->nameLen = 	strlen (s->name);
	s->searchPath = Lst_Init (FALSE);
	s->children = 	Lst_Init (FALSE);
	s->parents = 	Lst_Init (FALSE);
	s->sNum =   	sNum++;
	s->flags =  	0;

	(void)Lst_AtEnd (sufflist, (ClientData)s);
	/*
	 * Look for any existing transformations from or to this suffix.
	 * XXX: Only do this after a Suff_ClearSuffixes?
	 */
	Lst_ForEach (transforms, SuffRebuildGraph, (ClientData)s);
    } 
}

/*-
 *-----------------------------------------------------------------------
 * Suff_GetPath --
 *	Return the search path for the given suffix, if it's defined.
 *
 * Results:
 *	The searchPath for the desired suffix or NILLST if the suffix isn't
 *	defined.
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
Lst
Suff_GetPath (sname)
    char    	  *sname;
{
    LstNode   	  ln;
    Suff    	  *s;

    ln = Lst_Find (sufflist, (ClientData)sname, SuffSuffHasName);
    if (ln == NILLNODE) {
	return (NILLST);
    } else {
	s = (Suff *) Lst_Datum (ln);
	return (s->searchPath);
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_DoPaths --
 *	Extend the search paths for all suffixes to include the default
 *	search path.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The searchPath field of all the suffixes is extended by the
 *	directories in dirSearchPath. If paths were specified for the
 *	".h" suffix, the directories are stuffed into a global variable
 *	called ".INCLUDES" with each directory preceeded by a -I. The same
 *	is done for the ".a" suffix, except the variable is called
 *	".LIBS" and the flag is -L.
 *-----------------------------------------------------------------------
 */
void
Suff_DoPaths()
{
    register Suff   *s;
    register LstNode  ln;

    if (Lst_Open (sufflist) == FAILURE) {
	return;
    }
    /*
     * If they don't define a path for one of these variables, they still
     * must be defined, so we initialize them to the empty variable before
     * appending things in the loop below.
     */
    Var_Set (".INCLUDES", "", VAR_GLOBAL);
    Var_Set (".LIBS", "", VAR_GLOBAL);

    while ((ln = Lst_Next (sufflist)) != NILLNODE) {
	s = (Suff *) Lst_Datum (ln);
	if (!Lst_IsEmpty (s->searchPath)) {
	    Lst	  path;
#ifdef INCLUDES
	    if (s->flags & SUFF_INCLUDE) {
		Var_Append (".INCLUDES",
			 Dir_MakeFlags ("-I", s->searchPath),
			 VAR_GLOBAL);
	    }
#endif /* INCLUDES */
#ifdef LIBRARIES
	    if (s->flags & SUFF_LIBRARY) {
		Var_Append (".LIBS",
			 Dir_MakeFlags ("-L", s->searchPath),
			 VAR_GLOBAL);
	    }
#endif /* LIBRARIES */
	    path = Lst_Duplicate(dirSearchPath, Dir_CopyDir);
	    (void) Lst_Concat (s->searchPath, path, LST_CONCLINK);
	} else {
	    Lst_Destroy (s->searchPath, Dir_Destroy);
	    s->searchPath = Lst_Duplicate(dirSearchPath, Dir_CopyDir);
	}
    }
    Lst_Close (sufflist);
}

/*-
 *-----------------------------------------------------------------------
 * Suff_AddInclude --
 *	Add the given suffix as a type of file which gets included.
 *	Called from the parse module when a .INCLUDES line is parsed.
 *	The suffix must have already been defined.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The SUFF_INCLUDE bit is set in the suffix's flags field
 *
 *-----------------------------------------------------------------------
 */
void
Suff_AddInclude (sname)
    char	  *sname;     /* Name of suffix to mark */
{
    LstNode	  ln;
    Suff	  *s;

    ln = Lst_Find (sufflist, (ClientData)sname, SuffSuffHasName);
    if (ln != NILLNODE) {
	s = (Suff *) Lst_Datum (ln);
	s->flags |= SUFF_INCLUDE;
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_AddLib --
 *	Add the given suffix as a type of file which is a library.
 *	Called from the parse module when parsing a .LIBS line. The
 *	suffix must have been defined via .SUFFIXES before this is
 *	called.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The SUFF_LIBRARY bit is set in the suffix's flags field
 *
 *-----------------------------------------------------------------------
 */
void
Suff_AddLib (sname)
    char	  *sname;     /* Name of suffix to mark */
{
    LstNode	  ln;
    Suff	  *s;

    ln = Lst_Find (sufflist, (ClientData)sname, SuffSuffHasName);
    if (ln != NILLNODE) {
	s = (Suff *) Lst_Datum (ln);
	s->flags |= SUFF_LIBRARY;
    }
}

 	  /********** Implicit Source Search Functions *********/
/*
 * A structure for passing more than one argument to the Lst-library-invoked
 * function...
 */
typedef struct {
    Lst            l;
    Src            *s;
} LstSrc;

/*-
 *-----------------------------------------------------------------------
 * SuffAddSrc  --
 *	Add a suffix as a Src structure to the given list with its parent
 *	being the given Src structure. If the suffix is the null suffix,
 *	the prefix is used unaltered as the file name in the Src structure.
 *
 * Results:
 *	always returns 0
 *
 * Side Effects:
 *	A Src structure is created and tacked onto the end of the list
 *-----------------------------------------------------------------------
 */
static int
SuffAddSrc (s, ls)
    Suff	  *s;	  /* suffix for which to create a Src structure */
    LstSrc       *ls;	  /* list and parent for the new Src */
{
    Src           *s2;	  /* new Src structure */

    if (s->flags & SUFF_NULL) {
	/*
	 * If the suffix has been marked as the NULL suffix, also create a Src
	 * structure for a file with no suffix attached. Two birds, and all
	 * that...
	 */
	s2 = (Src *) malloc (sizeof (Src));
	s2->file =  	Str_New(pref);
	s2->parent = 	ls->s;
	s2->node =  	NILGNODE;
	s2->suff =  	s;
	(void)Lst_AtEnd (ls->l, (ClientData)s2);
    }
    s2 = (Src *) malloc (sizeof (Src));
    s2->file = 	    Str_Concat (pref, s->name, 0);
    s2->parent =    ls->s;
    s2->node = 	    NILGNODE;
    s2->suff = 	    s;
    (void)Lst_AtEnd (ls->l, (ClientData)s2);
}

/*-
 *-----------------------------------------------------------------------
 * SuffAddLevel  --
 *	Add all the children of targ as Src structures to the given list
 *
 * Results:
 *	None
 *
 * Side Effects:
 * 	Lots of structures are created and added to the list
 *-----------------------------------------------------------------------
 */
static void
SuffAddLevel (l, targ)
    Lst            l;		/* list to which to add the new level */
    Src            *targ;	/* Src structure to use as the parent */
{
    LstSrc         ls;

    ls.s = targ;
    ls.l = l;

    Lst_ForEach (targ->suff->children, SuffAddSrc, (ClientData)&ls);
}

/*-
 *----------------------------------------------------------------------
 * SuffFreeSrc --
 *	Free all memory associated with a Src structure
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	The memory is free'd.
 *----------------------------------------------------------------------
 */
static void
SuffFreeSrc (s)
    Src            *s;
{
    free ((Address)s->file);
    free ((Address)s);
}

/*-
 *-----------------------------------------------------------------------
 * SuffFindThem --
 *	Find the first existing file/target in the list srcs
 *
 * Results:
 *	The lowest structure in the chain of transformations
 *
 * Side Effects:
 *	None
 *-----------------------------------------------------------------------
 */
static Src *
SuffFindThem (srcs)
    Lst            srcs;	/* list of Src structures to search through */
{
    Src            *s;		/* current Src */
    Src		   *rs;		/* returned Src */

    rs = (Src *) NIL;

    while (!Lst_IsEmpty (srcs)) {
	s = (Src *) Lst_DeQueue (srcs);

	if (DEBUG(SUFF)) {
	    printf ("\ttrying %s...", s->file);
	}
	/*
	 * A file is considered to exist if either a node exists in the
	 * graph for it or the file actually exists.
	 */
	if ((Targ_FindNode(s->file, TARG_NOCREATE) != NILGNODE) ||
	    (Dir_FindFile (s->file, s->suff->searchPath) != (char *) NULL))
	{
	    if (DEBUG(SUFF)) {
		printf ("got it\n");
	    }
	    rs = s;
	    break;
	} else {
	    if (DEBUG(SUFF)) {
		printf ("not there\n");
	    }
	    SuffAddLevel (srcs, s);
	}
    }
    Lst_Destroy (srcs, SuffFreeSrc);
    return (rs);
}

/*-
 *-----------------------------------------------------------------------
 * SuffFindCmds --
 *	See if any of the children of the target in the Src structure is
 *	one from which the target can be transformed. If there is one,
 *	a Src structure is put together for it and returned.
 *
 * Results:
 *	The Src structure of the "winning" child, or NIL if no such beast.
 *
 * Side Effects:
 *	A Src structure may be allocated.
 *
 *-----------------------------------------------------------------------
 */
static Src *
SuffFindCmds (targ)
    Src	    	  	*targ;	/* Src structure to play with */
{
    LstNode 	  	ln; 	/* General-purpose list node */
    register GNode	*t, 	/* Target GNode */
	    	  	*s; 	/* Source GNode */
    int	    	  	prefLen;/* The length of the defined prefix */
    Suff    	  	*suff;	/* Suffix on matching beastie */
    Src	    	  	*ret;	/* Return value */
    char    	  	*cp;

    t = targ->node;
    (void) Lst_Open (t->children);
    prefLen = strlen (pref);

    while ((ln = Lst_Next (t->children)) != NILLNODE) {
	s = (GNode *)Lst_Datum (ln);

	cp = rindex (s->name, '/');
	if (cp == (char *)NULL) {
	    cp = s->name;
	} else {
	    cp++;
	}
	if (strncmp (cp, pref, prefLen) == 0) {
	    /*
	     * The node matches the prefix ok, see if it has a known
	     * suffix.
	     */
	    ln = Lst_Find (sufflist, (ClientData)&cp[prefLen],
			   SuffSuffHasName);
	    if (ln != NILLNODE) {
		/*
		 * It even has a known suffix, see if there's a transformation
		 * defined between the node's suffix and the target's suffix
		 */
		suff = (Suff *)Lst_Datum (ln);

		if (Lst_Member (suff->parents,
				(ClientData)targ->suff) != NILLNODE)
		{
		    /*
		     * Hot Damn! Create a new Src structure to describe
		     * this transformation (making sure to duplicate the
		     * source node's name so Suff_FindDeps can free it
		     * again (ick)), and return the new structure.
		     */
		    ret = (Src *)malloc (sizeof(Src));
		    ret->file = Str_New(s->name);
		    ret->suff = suff;
		    ret->parent = targ;
		    ret->node = s;
		    if (DEBUG(SUFF)) {
			printf ("\tusing existing source %s\n", s->name);
		    }
		    return (ret);
		}
	    }
	}
    }
    Lst_Close (t->children);
    return ((Src *)NIL);
}

/*-
 *-----------------------------------------------------------------------
 * SuffExpandChildren --
 *	Expand the names of any children of a given node that contain
 *	variable invocations or file wildcards into actual targets.
 *
 * Results:
 *	=== 0 (continue)
 *
 * Side Effects:
 *	The expanded node is removed from the parent's list of children,
 *	and the parent's unmade counter is decremented, but other nodes
 * 	may be added.
 *
 *-----------------------------------------------------------------------
 */
static int
SuffExpandChildren(cgn, pgn)
    GNode   	*cgn;	    /* Child to examine */
    GNode   	*pgn;	    /* Parent node being processed */
{
    GNode	*gn;	    /* New source 8) */
    LstNode	ln; 	    /* List element for old source */
    char	*cp;	    /* Expanded value */

    /*
     * First do variable expansion -- this takes precedence over
     * wildcard expansion. If the result contains wildcards, they'll be gotten
     * to later since the resulting words are tacked on to the end of
     * the children list.
     */
    if (index(cgn->name, '$') != (char *)NULL) {
	char	**argv;	    /* Array of new sources */
	int 	argc;	    /* Number of new sources */
	int 	i;  	    /* Index of current new source */

	if (DEBUG(SUFF)) {
	    printf("Expanding \"%s\"...", cgn->name);
	}
	cp = Var_Subst(cgn->name, pgn, TRUE);

	if (cp != (char *)NULL) {
	    argv = Str_BreakString(cp, " \t", "", &argc);
	    for (i = 1; i < argc; i++) {
		if (DEBUG(SUFF)) {
		    printf("%s...", argv[i]);
		}
		gn = Targ_FindNode(argv[i], TARG_CREATE);
		/*
		 * If gn isn't already a child of the parent, make it so and
		 * up the parent's count of unmade children.
		 */
		if (Lst_Member(pgn->children, (ClientData)gn) == NILLNODE) {
		    (void)Lst_AtEnd(pgn->children, (ClientData)gn);
		    (void)Lst_AtEnd(gn->parents, (ClientData)pgn);
		    pgn->unmade++;
		}
	    }
	    /*
	     * Free the vector and the result
	     */
	    free((char *)argv);
	    free((char *)cp);
	}
	/*
	 * Now the source is expanded, remove it from the list of children to
	 * keep it from being processed.
	 */
	ln = Lst_Member(pgn->children, (ClientData)cgn);
	pgn->unmade--;
	Lst_Remove(pgn->children, ln);
	if (DEBUG(SUFF)) {
	    printf("\n");
	}
    } else if (Dir_HasWildcards(cgn->name)) {
	Lst 	exp;	    /* List of expansions */
	Lst 	path;	    /* Search path along which to expand */

	/*
	 * Find a path along which to expand the word.
	 *
	 * If the word has a known suffix, use that path.
	 * If it has no known suffix and we're allowed to use the null
	 *   suffix, use its path.
	 * Else use the default system search path.
	 */
	cp = cgn->name + strlen(cgn->name);
	ln = Lst_Find(sufflist, (ClientData)cp, SuffSuffIsSuffixP);

	if (DEBUG(SUFF)) {
	    printf("Wildcard expanding \"%s\"...", cgn->name);
	}
	
	if (ln != NILLNODE) {
	    Suff    *s = (Suff *)Lst_Datum(ln);

	    if (DEBUG(SUFF)) {
		printf("suffix is \"%s\"...", s->name);
	    }
	    path = s->searchPath;
	} else if ((suffNull != (Suff *)NULL) && !amMake) {
	    if (DEBUG(SUFF)) {
		printf("using .NULL(%s)...", suffNull->name);
	    }
	    path = suffNull->searchPath;
	} else {
	    /*
	     * Use default search path
	     */
	    path = dirSearchPath;
	}

	/*
	 * Expand the word along the chosen path
	 */
	exp = Dir_Expand(cgn->name, path);

	while (!Lst_IsEmpty(exp)) {
	    /*
	     * Fetch next expansion off the list and find its GNode
	     */
	    cp = (char *)Lst_DeQueue(exp);

	    if (DEBUG(SUFF)) {
		printf("%s...", cp);
	    }
	    gn = Targ_FindNode(cp, TARG_CREATE);

	    /*
	     * If gn isn't already a child of the parent, make it so and
	     * up the parent's count of unmade children.
	     */
	    if (Lst_Member(pgn->children, (ClientData)gn) == NILLNODE) {
		(void)Lst_AtEnd(pgn->children, (ClientData)gn);
		(void)Lst_AtEnd(gn->parents, (ClientData)pgn);
		pgn->unmade++;
	    }
	}

	/*
	 * Nuke what's left of the list
	 */
	Lst_Destroy(exp, NOFREE);
    
	/*
	 * Now the source is expanded, remove it from the list of children to
	 * keep it from being processed.
	 */
	ln = Lst_Member(pgn->children, (ClientData)cgn);
	pgn->unmade--;
	Lst_Remove(pgn->children, ln);
	if (DEBUG(SUFF)) {
	    printf("\n");
	}
    }

    return(0);
}

/*-
 *-----------------------------------------------------------------------
 * Suff_FindDeps  --
 *	Find implicit sources for the target described by the graph node
 *	gn
 *
 * Results:
 *	Nothing.
 *
 * Side Effects:
 *	Nodes are added to the graph below the passed-in node. The nodes
 *	are marked to have their IMPSRC variable filled in. The
 *	PREFIX variable is set for the given node and all its
 *	implied children.
 *
 * Notes:
 *	The path found by this target is the shortest path in the
 *	transformation graph, which may pass through non-existent targets,
 *	to an existing target. The search continues on all paths from the
 *	root suffix until a file is found. I.e. if there's a path
 *	.o -> .c -> .l -> .l,v from the root and the .l,v file exists but
 *	the .c and .l files don't, the search will branch out in
 *	all directions from .o and again from all the nodes on the
 *	next level until the .l,v node is encountered.
 *
 *-----------------------------------------------------------------------
 */
void
Suff_FindDeps (gn)
    GNode         *gn;	      	/* node we're dealing with */
{
    Lst           srcs;	      	/* list on which to place Src structures */
    LstNode       transLn,   	/* node in transforms list */
                  ln;	      	/* node in ??? list */
    Suff          *s;	      	/* suffix on gn */
    Suff	  *as,	      	/* suffix for archive, if any */
		  *ms;	      	/* suffix for member of archive */
    Src           *src,	      	/* source Src structure when building chain
			      	 * of dependencies */
                  *targ;    	/* target Src structure for same purpose */
    GNode         *bottom,    	/* bottom-most graph node in chain */
                  *transGn;   	/* transform graph node */
    register GNode *sGn,     	/* graph node of source (from src) */
                  *tGn;	      	/* graph node of target (from targ) */
    char          *eopref;    	/* pointer to end of prefix */
    char	  *eoname;    	/* pointer to end of gn's name */
    char	  *eoarch;    	/* end of archive name */
    char	  *memPath;   	/* path to the archive member */
    char	  *name;      	/* The target name */
    char	  *cp;	      	/* useful character pointer */

    if (gn->type & OP_MEMBER) {
	/*
	 * If the node is a member of an archive, we already processed the
	 * beast for its parent node, so no point in doing it twice.
	 */
	return;
    }
    
    /*
     * Determine the suffix of the ultimate target.
     * At the end of this code segment, 'eopref' will point to the end
     * of the prefix (in gn->name), 'eoname' is the end of the suffix
     * plus one and 's' is the description of the file's suffix (its Suff
     * structure). The 'path' field of gn will be filled in, if the file
     * exists, and the TARGET variable for the node will have been set
     * appropriately. If the node is an archive, 'as' will contain the
     * suffix of the archive file.
     */
    (void)Lst_Open (sufflist);

    if (DEBUG(SUFF)) {
	printf ("Suff_FindDeps (%s)\n", gn->name);
    }
    
    if (gn->type & OP_ARCHV) {
	/*
	 * The node is an archive(member) pair. so we must find a
	 * suffix for both of them. The archive's suffix is tracked in
	 * the 'as' pointer while the member's is in 'ms'. Only when both are
	 * set may we break out of the loop.
	 */
	eoarch = index (gn->name, '(');
	eoname = index (eoarch, ')');

	name = eoarch + 1;

	ms = as = (Suff *) NULL;

	*eoname = '\0';	  /* Nuke parentheses during suffix search */
	*eoarch = '\0';	  /* So a suffix can be found */

	while ((ln = Lst_Next (sufflist)) != NILLNODE) {
	    s = (Suff *) Lst_Datum (ln);
	    if (SuffSuffIsSuffix (s, eoarch) != (char *) NULL) {
		as = s;
		if (ms != (Suff *) NULL) {
		    break;
		}
	    }
	    if (ms == (Suff *) NULL &&
		(eopref = SuffSuffIsSuffix (s, eoname)) != (char *) NULL)
	    {
		eopref++;
		ms = s;
		if (as != (Suff *) NULL) {
		    break;
		}
	    }
	}

	/*
	 * In the past, we sought both the member and the archive on their
	 * respective search paths. Since we no longer search for targets
	 * on their paths, and we assume that both the archive and the
	 * member are targets (that's why they're used in this way), we search
	 * for neither of them.
	 */
	Var_Set (MEMBER, name, gn);
	Var_Set (TARGET, name, gn);

	Var_Set (ARCHIVE, gn->name, gn);

	/*
	 * Replace the opening paren, but leave the closing one nuked for
	 * later use.
	 */
	*eoarch = '(';
	s = ms;
    } else if (gn->type & OP_LIB) {
	/*
	 * If the node is a library, it is the arch module's job to find it
	 * and set the TARGET variable accordingly. We merely provide the
	 * search path, assuming all libraries end in ".a" (if the suffix
	 * hasn't been defined, there's nothing we can do for it, so we just
	 * set the TARGET variable to the node's name in order to give it a
	 * value).
	 */
	ln = Lst_Find (sufflist, (ClientData)LIBSUFF, SuffSuffHasName);
	if (ln != NILLNODE) {
	    s = (Suff *) Lst_Datum (ln);
	    Arch_FindLib (gn, s->searchPath);
	} else {
	    Var_Set (TARGET, gn->name, gn);
	}
	name = Var_Value (TARGET, gn);
	s = (Suff *) NULL;
    } else {
	/*
	 * The node is a regular file. Find its suffix and then find the
	 * file on the search path, filling in the TARGET context variable
	 * and the 'path' field of the node appropriately.
	 */
	eoname = gn->name + strlen (gn->name);
	name = gn->name;

	ms = (Suff *) NULL;
	while ((ln = Lst_Next (sufflist)) != NILLNODE) {
	    s = (Suff *) Lst_Datum (ln);
	    if (eopref = SuffSuffIsSuffix (s, eoname)) {
		eopref++;
		ms = s;
		break;
	    }
	}

	/*
	 * XXX: This searching could be put off until later (when we know for
	 * certain if it's a target or a source)...
	 */
	if (ms != (Suff *) NULL) {
	    gn->path = Dir_FindFile (name, ms->searchPath);
	} else if (!amMake && (suffNull != (Suff *)NULL)) {
	    /*
	     * If there is a null suffix, try and find the file on its
	     * search path, setting the suffix used to be the null suffix.
	     * Also need to set the end-of-prefix pointer to the end of the
	     * name...
	     */
	    ms = suffNull;
	    if (DEBUG(SUFF)) {
		printf("\tNo known suffix on %s. Using Null suffix\n",
			 gn->name);
	    }
	    eopref = eoname;
	    gn->path = Dir_FindFile (name, suffNull->searchPath);
	}
	
	if (gn->path != (char *)NULL) {
	    Var_Set (TARGET, gn->path, gn);
	} else {
	    Var_Set (TARGET, gn->name, gn);
	}

	s = ms;
    }

    Lst_Close (sufflist);

    if ((s != (Suff *)NULL) && (s->flags & SUFF_LIBRARY)) {
	/*
	 * If the suffix indicates that the target is a library,
	 * mark that in the node's type field
	 */
	gn->type |= OP_LIB;
    }
    
    /*
     * Need the prefix both for the $(*) variable and the forming of new
     * Src structures. Note that the directory-part of the target name is
     * stripped off.
     */
    cp = rindex (name, '/');
    if (cp == (char *) NULL) {
	cp = name;
    } else {
	cp += 1;
    }
    
    pref = Str_New (cp);

    if (s != (Suff *) NULL) {
	pref[(eopref - cp) ] = '\0';
    }

    Var_Set (PREFIX, pref, gn);

    /*
     * Now we've got the important local variables set, expand any sources
     * that contain variables or wildcards in their names still. This needs to
     * be done now so we can be sure to find the proper transformation rules,
     * in certain cases. Besides, now is as good a time as any...
     * NOTE: Can't use Lst_ForEach b/c SuffExpandChildren will be removing
     * elements from the list, something Lst_ForEach isn't happy with.
     */
    Lst_Open(gn->children);
    while((ln = Lst_Next(gn->children)) != NILLNODE) {
	SuffExpandChildren((GNode *)Lst_Datum(ln), gn);
    }
    Lst_Close(gn->children);
    
    /*
     * If, at this point, there is no known suffix for the file, we restore
     * any characters we nuked and call it a day.
     */
    if (s == (Suff *) NULL) {
	if (DEBUG(SUFF)) {
	    printf ("\tNo valid suffix on %s\n", gn->name);
	}

	if (gn->type & OP_ARCHV) {
	    *eoname = ')';
	}
	free (pref);
	if (!OP_NOP(gn->type) && (!Lst_IsEmpty(gn->children) ||
				  !Lst_IsEmpty(gn->commands)))
	{
	    /*
	     * Actually a target (on the lhs of a dependency operator AND
	     * has either children or commands) -- put only its name in the
	     * .TARGET variable. Also set its path to its name -- that's the
	     * only place we're allowed to find it.
	     */
	    Var_Set(TARGET, gn->name, gn);
	    gn->path = gn->name;
	}
	return;
    }


    if (gn->type & OP_ARCHV) {
	/*
	 * If the node we're on is an archive node, we have to be tricky.
	 * The problem comes because there is sometimes more than one path
	 * down from an archive suffix into the graph, (e.g. if you have two
	 * specifications lib.a(fish.o) and lib.a(fish.po), you will go down
	 * the suffix graph to either .o or .po both times, yet we *know*
	 * which suffix we want -- it's in ()s) so we have to forcibly link the
	 * two nodes together.
	 */
	GNode	  *mem;	    /* Graph Node for the member */

	bottom = mem = Targ_FindNode (name, TARG_CREATE);
	(void) Lst_AtEnd (gn->children, (ClientData)mem);
	(void) Lst_AtEnd (mem->parents, (ClientData)gn);
	gn->unmade += 1;

	transLn = Lst_Find (transforms,
			    (ClientData)Str_Concat (ms->name, as->name, 0),
			    SuffGNHasName);

	if (transLn != NILLNODE) {
	    /*
	     * If there's a decent transformation from the member's suffix
	     * to the archive's apply it to the archive(member) target.
	     */
	    if (DEBUG(SUFF)) {
		printf ("\tapplying %s -> %s to \"%s\"\n", ms->name,
			  as->name, mem->name);
	    }
	    transGn = (GNode *) Lst_Datum (transLn);
	
	    (void) Lst_Concat (gn->commands, transGn->commands, LST_CONCNEW);
	    gn->type |= transGn->type & ~(OP_OPMASK|OP_TRANSFORM);
	} else if (DEBUG(SUFF)) {
	    printf("\tno transformation from %s -> %s\n", ms->name, as->name);
	}

	(void)Lst_AtEnd (mem->iParents, (ClientData)gn);

	/*
	 * We need to find out NOW if the member is a target or a source only
	 * so we can set our own .TARGET variable correctly, so recurse to set
	 * the member's .TARGET variable properly and copy that in.
	 */
	Suff_FindDeps(mem);
	Var_Set(TARGET, Var_Value(TARGET, mem), gn);
	
	/*
	 * Make sure this node isn't seen as being only a source. This allows
	 * the user to not define a transformation from the member suffix to
	 * the archive suffix w/o PMake choking b/c it can't figure out how
	 * to make the archive(member) target.
	 */
	if (OP_NOP(gn->type)) {
	    gn->type |= OP_DEPENDS;
	}

	/*
	 * Flag that the member node is a member node so we can consult the
	 * archive if it doesn't exist.
	 */
	mem->type |= OP_MEMBER;
	*eoname = ')';
    } else {
	srcs = Lst_Init (FALSE);
	
	/*
	 * We make up a Src structure for the ultimate target so the loop finds
	 * it via parent links and we get back where we started.
	 */
	targ = (Src *) malloc (sizeof (Src));
	targ->file = Str_New (name);
	targ->suff = s;
	
	targ->node = gn;
	targ->parent = (Src *) NIL;
	
	src = (Src *)NIL;
	
	if (!Lst_IsEmpty (gn->children)) {
	    /*
	     * If the node has some explicit sources, see if we can transform
	     * to the current target from one of them.
	     * If not, we'll look for an implied source, otherwise
	     * we'll just use the one we've got.
	     */
	    src = SuffFindCmds (targ);
	}
	
	if (src == (Src *)NIL) {
	    /*
	     * Nothing explicit fits -- add the initial level and then pass
	     * the buck to SuffFindThem.
	     */
	    SuffAddLevel (srcs, targ);
	    src = SuffFindThem (srcs);
	}
	
	if (src == (Src *) NIL) {
	    /*
	     * Didn't find an implied source -- clean up after ourselves.
	     */
	    free ((Address)pref);
	    free ((Address)targ->file);
	    free ((Address)targ);
	    if (!OP_NOP(gn->type) && (!Lst_IsEmpty(gn->children) ||
				      !Lst_IsEmpty(gn->commands)))
	    {
		/*
		 * Actually a target (on the lhs of a dependency operator AND
		 * has either children or commands) -- put only its name in the
		 * .TARGET variable. Also set its path to its name -- that's
		 * the only place we're allowed to find it.
		 */
		Var_Set(TARGET, gn->name, gn);
		gn->path = gn->name;
	    }
	    return;
	}
	if (src->node == NILGNODE) {
	    sGn = bottom = Targ_FindNode (src->file, TARG_CREATE);
	} else {
	    sGn = bottom = src->node;
	}
	
	/*
	 * We now have a list of Src structures headed by 'src' and linked via
	 * their 'parent' pointers. What we do next is create links between
	 * source and target nodes (which may or may not have been created)
	 * and set the necessary local variables in each target. The
	 * commands for each target are set from the commands of the
	 * transformation rule used to get from the src suffix to the targ
	 * suffix. Note that this causes the commands list of the original
	 * node, gn, to be replaced by the commands of the final
	 * transformation rule. Also, the unmade field of gn is incremented.
	 * Etc. 
	 */
	while (src->parent != (Src *) NIL) {
	    targ = src->parent;
	    
	    /*
	     * If we haven't already created the node, do it now 
	     */
	    if (targ->node != NILGNODE) {
		tGn = targ->node;
	    } else {
		tGn = Targ_FindNode (targ->file, TARG_CREATE);
	    }
	    
	    /*
	     * Form the proper links and update the counters in the target node
	     */
	    if (Lst_Member (tGn->children, (ClientData)sGn) == NILLNODE) {
		(void)Lst_AtEnd (tGn->children, (ClientData)sGn);
		(void)Lst_AtEnd (sGn->parents, (ClientData)tGn);
		tGn->unmade += 1;
	    }
	    
	    /*
	     * Now find the transformation rule and stuff its commands into the
	     * target node, destroying the old commands list. 
	     */
	    transLn = Lst_Find (transforms,
				(ClientData)Str_Concat (src->suff->name,
							targ->suff->name, 0),
				SuffGNHasName);
	    transGn = (GNode *) Lst_Datum (transLn);
	    
	    if (DEBUG(SUFF)) {
		printf ("\tapplying %s -> %s to \"%s\"\n", src->suff->name,
			  targ->suff->name, sGn->name);
	    }
	    /*
	     * While we look for implicit sources this way, we only apply the
	     * commands that come from the rule if the target has no commands
	     * of its own.
	     */
	    if (Lst_IsEmpty(tGn->commands)) {
		(void) Lst_Concat (tGn->commands, transGn->commands,
				   LST_CONCNEW);
	    }
	    tGn->type |= transGn->type & ~OP_OPMASK;
	    
	    /*
	     * At this point, two context variables need to be set. One of
	     * them, $(*) we can set now, since the prefix is the prefix. The
	     * other must wait until the source node actually has a path
	     * installed. Because of this, we simply mark the source so in
	     * either of the update functions in make.c, the $(<) variable is
	     * set for the parent that needs it.
	     */
	    (void) Lst_AtEnd (sGn->iParents, (ClientData)tGn);
	    if (tGn != gn) {
		Var_Set (PREFIX, pref, tGn);
	    }
	    
	    /*
	     * We now have no need for the src structure 
	     */
	    free ((Address)src->file);
	    free ((Address)src);
	    
	    src = targ;
	    sGn = tGn;
	}
	free (pref);

	/*
	 * We know the node is actually a target, since we gave it sources,
	 * so put its name in the .TARGET variable, not its path.
	 */
	Var_Set(TARGET, gn->name, gn);
	gn->path = gn->name;
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_SetNull --
 *	Define which suffix is the null suffix.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	'suffNull' is altered.
 *
 * Notes:
 *	Need to handle the changing of the null suffix gracefully so the
 *	old transformation rules don't just go away.
 *
 *-----------------------------------------------------------------------
 */
void
Suff_SetNull(name)
    char    *name;	    /* Name of null suffix */
{
    Suff    *s;
    LstNode ln;

    ln = Lst_Find(sufflist, (ClientData)name, SuffSuffHasName);
    if (ln != NILLNODE) {
	s = (Suff *)Lst_Datum(ln);
	if (suffNull != (Suff *)NULL) {
	    suffNull->flags &= ~SUFF_NULL;
	}
	s->flags |= SUFF_NULL;
	/*
	 * XXX: Here's where the transformation mangling would take place
	 */
	suffNull = s;
    } else {
	Parse_Error (PARSE_WARNING, "Desired null suffix %s not defined.",
		     name);
    }
}

/*-
 *-----------------------------------------------------------------------
 * Suff_Init --
 *	Initialize suffixes module
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Many
 *-----------------------------------------------------------------------
 */
void
Suff_Init ()
{
    sufflist = Lst_Init (FALSE);
    transforms = Lst_Init (FALSE);

    sNum = 0;
    suffNull = (Suff *)NULL;
}

/********************* DEBUGGING FUNCTIONS **********************/

static int SuffPrintName(s) Suff *s; {printf ("%s ", s->name); return (0);}

static int
SuffPrintSuff (s)
    Suff    *s;
{
    int	    flags;
    int	    flag;

    printf ("# %s", s->name);
    
    flags = s->flags;
    if (flags) {
	fputs (" (", stdout);
	while (flags) {
	    flag = 1 << (ffs(flags) - 1);
	    flags &= ~flag;
	    switch (flag) {
		case SUFF_NULL:
		    printf ("NULL");
		    break;
		case SUFF_INCLUDE:
		    printf ("INCLUDE");
		    break;
		case SUFF_LIBRARY:
		    printf ("LIBRARY");
		    break;
	    }
	    putc(flags ? '|' : ')', stdout);
	}
    }
    putc ('\n', stdout);
    printf ("#\tTo: ");
    Lst_ForEach (s->parents, SuffPrintName, (ClientData)0);
    putc ('\n', stdout);
    printf ("#\tFrom: ");
    Lst_ForEach (s->children, SuffPrintName, (ClientData)0);
    putc ('\n', stdout);
    printf ("#\tSearch Path: ");
    Dir_PrintPath (s->searchPath);
    putc ('\n', stdout);
    return (0);
}

static int
SuffPrintTrans (t)
    GNode   *t;
{
    extern int Targ_PrintCmd();

    printf ("%-16s: ", t->name);
    Targ_PrintType (t->type);
    putc ('\n', stdout);
    Lst_ForEach (t->commands, Targ_PrintCmd, (ClientData)0);
    putc ('\n', stdout);
    return(0);
}

Suff_PrintAll()
{
    printf ("#*** Suffixes:\n");
    Lst_ForEach (sufflist, SuffPrintSuff, (ClientData)0);

    printf ("#*** Transformations:\n");
    Lst_ForEach (transforms, SuffPrintTrans, (ClientData)0);
}

